from setuptools import setup, find_packages
from os import path

this_directory = path.abspath(path.dirname(__file__))

long_description = """

For full documentation please check [ML-Navigator](https://ki-labs.github.io/ML-Navigator/index.html)


[![Build Status](https://travis-ci.com/KI-labs/ML-Navigator.svg?token=qeb22MSpJyy4b3twFCDG&branch=master)](https://travis-ci.com/KI-labs/ML-Navigator)
[![License](https://img.shields.io/badge/License-Apache%202.0-blue.svg)](https://opensource.org/licenses/Apache-2.0)

# Concept

ML-Navigator is a tutorial-based Machine Learning framework. The main component of ML-Navigator is the flow.
A flow is a collection of compact methods/functions that can be stuck together with guidance texts. 

The flow functions as a map which shows the road from point A to point B. The guidance texts function as navigator 
instructions that help the user to figure out the next step after executing the current step.

Like the car navigator, the user is not forced to follow the path. At any point, the user can take a break to explore 
data, modify the features and make any necessary changes. The user can always come back to the main path which the flow 
defines.

> The flows are created by the community for the community.

# Introduction

> ML-Navigator is a free editable maps collection of the data science world.<br>

## Why ML-Navigator

ML-Navigator standardizes exchanging knowledge among data scientists. Junior data scientists
can learn and apply data science best-practice in a white-box mode. Senior data scientists can
automate many repetitive processes and share experience effectively. Enterprises can standardize
data science among different departments.

## Background
Data Science has been attracting smart people who have different backgrounds, 
experiences, and field knowledge. The transformation journey from other disciplines into 
data science is not a straight forward process. It is time and effort consuming based on 
the motivation, the background, and the industry where the data scientist wants to work.
The data science new joiners follow multiple paths to sharp their data science skills. 
Some of these paths are:<br>
- Online courses: Some E-learning platforms, e.g., LinkedIn-learning, 
provide practical courses to solve specific data science problems. 
Other platforms, such as Coursera, offer theory-based courses. In LinkedIn-learning 
platform alone, there are 400+ courses related to data science. Selecting the best 
courses among all of those numerous courses is a challenge for newbies. Moreover, 
the theory-based courses require solid mathematical knowledge, especially in 
calculus and linear algebra.
- Data Science online platforms: Such platforms, like Kaggle, offer playground or 
prize-based competitions. Junior data scientists can learn a lot by applying their 
knowledge and reading kernels, which data scientists write to share their experience. 
Poorly written code and lack of documentation can be frustrating for newbies who want to 
learn what happened behind the scenes.
- Manuals of well-known data science frameworks: There are many open-source frameworks 
which provide an industry-proven implementation of many methods that have been used by 
data scientists. Many of these frameworks don't share the same syntax. Data scientists 
may need to learn new syntax each time they switch to a new framework.
- Learning from senior data scientists: Onboarding junior data scientists may require 
time which senior data scientists don't always have.


## To whom is ML-Navigator
### Data science new joiners
A new joiner is a person who wants to move into data science from a different discipline.
A new joiner can also be a person who wants to be a part of the data team but not
 a full-time data scientist, e.g., developers with sufficient coding skills.
ML-Navigator provides the data science new joiners the path to analyze real data.
It helps the user to navigate through predefined flows, which are End-2-End data
 science pipelines. The user can load a specific flow and follow the instructions starting
 from reading data until training the model. 
The user can start with the most straightforward flow and later use more complicated 
flows to train accurate models if needed.

### Senior data scientists
Experienced data scientists may be interested in automating many processes that they follow
 frequently. They can build a flow for each specific problem type. 
 The flow can be created from scratch or by modifying or combining other flows. 
 They can share their flows with the community and exchange their experience with other data 
 scientists.


### ML-Navigator for enterprises
ML-navigator can standardize the data science experience in large enterprises.
Junior data scientists can be productive and efficient from the first day. The onboarding process
can be fast, concrete, but not abstracted.

Data scientists may use AutoML to produce multiple types of models as an alternative to digging deep in data and gaining 
new knowledge. AutoML can create a large number of models. 
However, it doesn't guarantee that the user gets the model that satisfies the quality requirements. 
It needs a long time for testing a wide range of hyperparameters values. 
Model reproducibility can be an issue when creating models using AutoML.

# How-to Guides
## How to install ML-Navigator

To install the ML-Navigator Package you need to have `Python 3.6`:

You can install ML-Navigator using the `pip` tool directly:

`pip install ML-Navigator`

To install the ML-Navigator Package from the Github repo:

1. clone the git repository:

    `$ git clone https://github.com/KI-labs/ML-Navigator.git`<br>
    `$ cd ML-Navigator`<br>
    

2. create a directory under the name "data" and move your data files to it e.g. "train.csv" and "test.csv"

3. create a virtual environment

    `$ pip install virtualenv`<br>
    `$ virtualenv venv`<br>
    `$ source /venv/bin/activate`
    
4. After setting up the virtual environment, you can install the package using pip command as follows:

   `$ pip install .`<br>
    
## File Structure
The structure of the directories looks like the following

````
.
├── LICENSE
├── setup.py
├── MANIFEST.in
├── data
│   ├── flow_0
│   ├── flow_1
│   ├── flow_2
│   └── flow_3
├── feature_engineering
│   ├── __init__.py
│   ├── feature_generator.py
│   └── test.py
├── flows
│   ├── __init__.py
│   ├── example.yaml
│   ├── flow_0.drawio
│   ├── flow_0.json
│   ├── flow_1.drawio
│   ├── flow_1.json
│   ├── flow_2.drawio
│   ├── flow_2.json
│   ├── flow_3.drawio
│   ├── flow_3.json
│   ├── flows.py
│   ├── utils.py
│   ├── text_helper.py
│   └── yaml_reader.py
├── images
│   ├── flow_0_record_middle_size.gif
│   └── logo.png
├── prediction
│   ├── __init__.py
│   └── model_predictor.py
├── preprocessing
│   ├── README.md
│   ├── __init__.py
│   ├── data_clean.py
│   ├── data_explorer.py
│   ├── data_science_help_functions.py
│   ├── data_transformer.py
│   ├── data_type_detector.py
│   ├── json_preprocessor.py
│   ├── test_loading_data.py
│   ├── test_preprocessing.py
│   └── utils.py
├── readme.md
├── requirements.txt
├── training
│   ├── __init__.py
│   ├── model_evaluator.py
│   ├── optimizer.py
│   ├── test_split.py
│   ├── test_training.py
│   ├── training.py
│   ├── utils.py
│   └── validator.py
├── tutorials
│   ├── flow_0.png
│   ├── flow_0.ipynb
│   ├── flow_1.ipynb
│   ├── flow_1.png
│   ├── flow_2.ipynb
│   ├── flow_2.png
│   ├── flow_3.ipynb
│   └── flow_3.png
├── venv
└── visualization
    ├── __init__.py
    └── visualization.py

````

# Tutorials

Create a directory under the name "data" inside the project root directory.

To run the tutorials, you can download the "train.csv" and "test.csv" datasets from Kaggle website:

    * inside `./data/flow_0` and `./data/flow_1` store the data from the [House Prices: Advanced Regression Techniques competition](https://www.kaggle.com/c/house-prices-advanced-regression-techniques)

    * inside `./data/flow_2` directory store the data from the: [TMDB Box Office Prediction competition](https://www.kaggle.com/c/tmdb-box-office-prediction)

Please check the tutorials directory: [Tutorials](https://github.com/KI-labs/ML-Navigator/tree/master/tutorials)

# Reference

For more information please check the documentation: [ML-Navigator](https://ki-labs.github.io/ML-Navigator/index.html)

## Modules

### flows

It is the core module of the ML-Navigator framework. It contains the flows that are defined as YAML files.
The Flow class has multiple methods that get their functionalities from calling other packages (self-implemented or external).

### preprocessing

It contains the main functions and classes that are used to prepare the data for further processing such as discovering
the type of data in each column, encode categorical features and scale numeric features.

## feature_engineering

It contains functions and classes to produce new features. For example, one-hot encoding.

## visualization

It contains functions and classes to produce graphs. For example, there are functions for comparing the statistical
 properties of different datasets, visualize the count of the missing values, and drawing histograms.

## training

It contains the functions and classes to train Machine Learning models. Currently, there are two regression models:
 Ridge linear regression (scikit-learn) and LightGBM.
 
### prediction

It contains the functions and classes to predict the target using the pre-trained models. Currently, all trained models are saved locally.

## Other directories

### logs

It contains the logs messages that are produced by different modules of the framework.

### models

It contains the trained models saved in pkl format.

### venv

The virtual environment that should be created by the user

### data

It contains the data, e.g. `train.csv` and `test.csv`.

## Files:

### requirements.txt

It contains a list of all packages that are required to run the framework.

## How to apply flows to your data

It is straightforward. You need to point to the location of your data and the name of the datasets after loading a particular flow.
Currently, the framework supports only reading CSV files. Here is an example:

````python
path = './data'
files_list = ['train.csv','test.csv']
````

## How to create a flow

### Introduction

For the first version, we have not prepared a tool for creating a flow in a cool way yet. However, this is one of the main focus in the near future.

The flow should contain two elements:

1. Visualization: I show a flow as a flowchart. I use a free online tool called draw.io to draw a chart. Feel free to use any other tool or method to visualize a flow.
You can use the `drawio` files which are provided in the flows modules to create a visualization of new flows.

2. guidance text: I use YAML files to define the guidance instructions for the flows. Currently, this method is not scalable, and it requires setting the steps manually.
In the future, the flows will be created using a user interface, and they will be saved in a database using a unique key for each flow.

Each method in the `Flows` class in the `flows/flow.py` has an ID. Each ID is defined as a string using the variable `function_id`. For example, the method `load_data` has `function_id = 0`.
In the `flows/flow_instructions_database.yaml` there are the instruction texts. The instruction texts are defined using an ID (integer). Each text has two variables:
* `function`: it describes the function or the method which is defined in `flows/flow.py`
* `guide`: it the guidance text that describe how to use the defined function or the method.

To build a flow you need to create a JSON file, e.g. `flow_0.json`, that map the `function_id` as a key and the ID of the guidance text that is defined in `flows/flow_instructions_database.yaml`.
<br>
<strong>IMPORTANT!!!!!</strong><br>
In the `flow_x.json`:<br>
`function_id` refers to the id of the current running function<br>
The ID of the guidance text in the `flow_instructions_database.yaml` refers to the function that should be executed after the current running function that has the `function_id`.

````json
{"function_id of running function": "the ID of the guidance text of the function that should be executed next"}
````

An example for the mapping:
````json
{"0": 1}
````

where `function_id = 0` refers to the `load_data` method and value 1 refers to the ID of the method `Encode categorical features` guidance text that should run after the `load_data` method.

### Example of creating the flow_0.json

````json
{
  "0": 1,
  "1": 2,
  "2": 3,
  "4": 1000
}
````

The translation of the JSON object is as follows:

````json
{
  "the current function: load the data": "the next function: Encode categorical features",
  "the current function: Encode categorical features": "the next function: Scale numeric features",
  "the current function: Scale numeric features": "the next function: Train a model",
  "the current function: Train a model": "the next function: Finish or noting which indicates the end of the flow"
}
````

### Creating new flows components

You can create your own method inside the `Flows` class in the `flows/flow.py` and assign a unique ID to it by defining the variable `function_id`.
Inside the `flow_instructions_database.yaml` you can create your own guidance text for already exiting methods or for the new methods. You should assign a unique ID for the 
new created guidance texts. Please include the `function` and `quide` keys to help other users understanding and find your guidance text easily.
The key `function` is optional but `quide` is required. You can create multiple new guidance texts for the same defined function but each guidance text should have a unique ID.

When creating a flow, the essential information that should be added at the end of each step is what the next step is. Adding an example, which shows how to perform
the next level and what are the required variables is beneficial to the user.

# Contribution

Your contributions are always welcome and appreciated. Following are the things you can do to contribute to this project.

## Report a bug
If you think you have encountered a bug, and I should know about it, feel free to report it [here](https://github.com/KI-labs/ML-Navigator/issues) and I will take care of it.

## Request a feature
You can also request for a feature [here](https://github.com/KI-labs/ML-Navigator/issues), and if it will viable, it will be picked for development.

## Create a pull request
It can't get better then this, your pull request will be appreciated by the community. You can get started by picking up any open issues from [here](https://github.com/KI-labs/ML-Navigator/issues) and make a pull request.

## create a new flow:
If you want to submit a flow, please provide the following in your pull request:
1. `flow_x.drawio` and `flow_x.png` where x is an integer that has not been given for other flows yet. Please check the flows module `./flows`<br>

2. `flow_x.json` where x is is an integer that has not been given for other flows yet and has the same value in `flow_x.drawio`<br>

3. `flow_x.ipynb` where x is is an integer that has not been given for other flows yet and has the same value in `flow_x.drawio` and in `flow_x.json`.<br>
The Jupyter Notebook `flow_x.ipynb` should work end-2-end without any errors.

Steps to create a pull request
------------------------------

1. Make a PR to master branch.
2. Comply with the best practices and guidelines.
3. It must pass all continuous integration checks and get positive reviews.
4. After this, changes will be merged.

# License

Copyright 2019 KI labs GmbH

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

with open('./requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name="ML-Navigator",
    version="0.0.31",
    python_requires='>=3.6',
    packages=find_packages(),
    install_requires=requirements,
    include_package_data=True,
    author="Maher Deeb",
    description="ML-Navigator is a tutorial-based Machine Learning framework. The main component of ML-Navigator is "
                "the flow. A flow is a collection of compact methods/functions that can be stuck together with guidance"
                " texts.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords="Machine Learning Data Science ML-Navigator tutorial-based data pipelines flows flow",
    url="https://www.ki-labs.com/",
    project_urls={
        "Bug Tracker": "https://github.com/KI-labs/ML-Navigator/issues",
        "Documentation": "https://ki-labs.github.io/ML-Navigator/index.html",
    },
    classifiers=[
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: Apache Software License'
    ]
)



import logging
import os

import numpy as np
from sklearn import linear_model

from training.model_evaluator import evaluate_model

logger = logging.getLogger(__name__)
formatting = (
    "%(asctime)s: %(levelname)s: File:%(filename)s Function:%(funcName)s Line:%(lineno)d "
    "message:%(message)s"
)
logging.basicConfig(
    filename=os.path.join(os.path.dirname(os.path.dirname(__file__)), "logs/logs.log"),
    level=logging.INFO,
    format=formatting,
)


def training_for_optimizing(alpha_i: float, x_train: np.array, y_train: np.array, x_test: np.array, y_test: np.array,
                            help_text: str) -> float:
    """Trainer

    This function trains the ridge linear regression model given a certain alpha.

    :param float alpha_i: A hyperparameter which used by the ridge linear regression to avoid over-fitting.
    :param np.array x_train: The values of the features which are used to train the model to predict the target y_train
    :param np.array y_train: The target values which are used to train the model.
    :param np.array x_test: The values of the features which are used to evaluate the model by predicting the target
            y_test
    :param np.array y_test: The target values which are used to evaluate the performance of the model based on the
            coefficient of determination R2
    :param str help_text: A string to show useful information about the training cross-validation method
    :return:
            - r2_linear - Coefficient of determination for a given alpha and testing dataset
    """

    linear_regression_model = linear_model.Ridge(alpha=alpha_i)
    model = linear_regression_model.fit(x_train, y_train)

    metrics_summary = evaluate_model(model,
                                     xs=[x_test],
                                     ys=[y_test],
                                     labels=[""],
                                     metrics=["r2_score"])

    r2_linear = metrics_summary['r2_score ']

    return r2_linear


def get_best_alpha_split(x_train: np.array, y_train: np.array, x_test: np.array, y_test: np.array) -> float:
    """ alpha optimizer for two datasets split

    This function finds the best alpha value based on the coefficient of determination.
    This function will be replaced by a native optimization method from other packages

    :param np.array x_train: The values of the features which are used to train the model to predict the target y_train
    :param np.array y_train: The target values which are used to train the model.
    :param np.array x_test: The values of the features which are used to evaluate the model by predicting the target
            y_test
    :param np.array y_test: The target values which are used to evaluate the performance of the model based on the
            coefficient of determination R2

    :return:
            best_alpha: The alpha value that maximize the coefficient of determination R2
    """

    help_text = "Testing dataset from splitting"

    # generate a list of (alpha, r2_score) pairs
    candidates = [(alpha, training_for_optimizing(alpha, x_train, y_train, x_test, y_test, help_text))
                  for alpha in np.logspace(-3, 3, 19)]

    best_alpha, _ = max(candidates, key=lambda x: x[1])

    logger.info("Alpha optimization for split cross-validation is finished")

    return best_alpha


def get_best_alpha_kfold(kfold, train_array: np.array, target: np.array):
    """ alpha optimizer for K-Fold cross validation

    :param kfold:
    :param train_array: The values of the features which are used to train the model to predict the target target
    :param target: The target values which are used to train the model.

    :return:
            best_alpha: The alpha value that maximize the coefficient of determination R2
    """

    # generate a list of (alpha, r2_score) pairs
    candidates = [
        (alpha,
         np.mean([training_for_optimizing(alpha,
                                          train_array[train],
                                          target[train],
                                          train_array[test],
                                          target[test],
                                          f"dataset kfold {idx}")
                  for idx, (train, test) in enumerate(kfold.split(train_array, target))])
         ) for alpha in np.logspace(-3, 3, 19)
    ]

    best_alpha, _ = max(candidates, key=lambda x: x[1])

    logger.info("Alpha optimization for Kfold cross-validation is finished")

    return best_alpha



from blessings import Terminal

# For colorful and beautiful formatted print()
term = Terminal()


class FlowInstructions:

    @staticmethod
    def read_data(specific_flow: int = 0,
                  specific_directory: str = "./data",
                  specific_file_list: str = "'train.csv','test.csv'"):
        print(term.bold(term.magenta("Please use the following function to read the data")))
        print(term.green_on_black("dataframe_dict, columns_set = flow.load_data(path : str, files_list : list)"))
        print(term.bold(term.magenta("For example: ") + term.green_on_black(f"path = {specific_directory}")))
        print(term.bold(term.magenta(
            "If your data is in a nested directory, it is better to os.path.join. For example:\n") + term.green_on_black(
            f"path = os.path.join('data', 'flow_{specific_flow}')")))
        print(term.bold(term.magenta("For example: ") + term.green_on_black(f"files_list = [{specific_file_list}]")))
        print(term.bold(term.magenta("The output is a dictionary that contains dataframes e.g.\n")))
        print(term.blue("dataframe_dict = {'train': train_dataframe,'test': test_dataframe}"))

    @staticmethod
    def encode_categorical_features():
        print(term.bold(term.magenta("If you have categorical features with string labels, Encode the categorical "
                                     "features by applying the following function:\n") + term.green_on_black(
            "dataframe_dict, columns_set = flow.encode_categorical_feature(dataframe_dict: dict)")))

    @staticmethod
    def scale_numeric_features():
        print(term.bold(term.magenta("If you have numeric features, it is a good idea to normalize numeric features." +
                                     "Use the following function for feature normalization:\n") +
                        term.green_on_black(
                            " dataframe_dict, columns_set = flow.scale_data (dataframe_dict:"
                            " dict, ignore_columns: list)")))
        print(term.bold(term.magenta("For example: ") + term.green_on_black("ignore_columns = ['id', 'target']")))

    @staticmethod
    def train_a_model():
        print(term.bold(term.magenta("Your features are ready to train the model: ")))
        print(term.bold(term.magenta("You can start training the model by applying the following function: ")))
        print(term.green_on_black("model_index_list, save_models_dir, y_test = flow.training(parameters)"))
        print('parameters = {\n'
              '     "data": {\n'
              '         "train": {"features": train_dataframe, "target": train_target}, \n'
              '         "valid": {"features": valid_dataframe, "target": valid_target}, \n'
              '         "test": {"features": test_dataframe, "target": test_target}, \n'
              '     }, \n'
              '     "split": {\n'
              '         "method": "split",\n'
              '         "split_ratios": 0.2\n'
              '         }, \n '
              '     "model": {"type": "Ridge linear regression",\n'
              '               "hyperparameters": {"alpha": 1,  # alpha:optimize}\n'
              '             }, \n '
              '     "metrics": ["r2_score", "mean_squared_error"]\n'
              '}')

    @staticmethod
    def one_hot_encoding():
        print(term.bold(term.magenta(
            "You have categorical features. Apply one-hot encoding to the categorical features by applying the"
            " following function:\n") + term.green_on_black('dataframe_dict, columns_set = flow.features_encoding('
                                                            '"one-hot", dataframe_dict: dict, reference: str,'
                                                            ' ignore_columns: list, class_number_range = [3, 50])')))
        print(term.bold(term.magenta("Since one-hot encoding can produce a lot of features, class_number_range will "
                                     "limit the encoding process only for features which have between 3 and 49"
                                     " unique values.")))
        print(term.bold(term.magenta("If you are solving a classification problem, you should exclude the target from "
                                     "the one - hot encoding process by defining the ignore_columns\n ")
                        + term.green_on_black(" ignore_columns = [ < your target / label >]\n ")
                        + term.magenta("You can add more columns to the ignore_columns list to ignore")))

    @staticmethod
    def flatten_JSON_data():
        print(term.bold(term.magenta(
            "You have JSON nested data inside the dataframe columns. "
            "Flatten the nested JSON data by applying the following function:\n") + term.green_on_black(
            "dataframe_dict, columns_set= flow.flatten_json_data(dataframe_dict)")))

    @staticmethod
    def drop_high_corrected_columns_but_keep_one():
        print(term.bold(term.magenta(
            "If some features are highly correlated, they do not provide more information about the target prediction. "
            "It is a good idea to drop such features but keep one:\n") + term.green_on_black(
            "dataframe_dict, columns_set = flow.drop_correlated_columns(dataframe_dict: dict, ignore_columns: list)") +
                        term.magenta("An example of the ignore_columns list: \n")
                        + term.green_on_black(" ignore_columns = [target]\n")))

    @staticmethod
    def delete_features_with_constant_values():
        print(term.bold(term.magenta(
            "If some features have in all rows the same value, they have no influence on the target prediction. "
            "It is a good idea to delete such features:\n") + term.green_on_black(
            "dataframe_dict, columns_set = flow.drop_columns_constant_values("
            "dataframe_dict: dict, ignore_columns: list)") +
                        term.magenta("An example of the ignore_columns list: \n")
                        + term.green_on_black(" ignore_columns = [target]\n")))

    @staticmethod
    def explore_data():
        print(term.bold(term.magenta("If you want to explore the data you can run one of the following functions: ")))
        print(term.bold(term.magenta("1 . ") + term.green_on_black(
            "flow.exploring_data(dataframe_dict: dict, key_i: str)")))
        print(term.bold(term.magenta("For example: ") + term.green_on_black(
            "flow.exploring_data(dataframe_dict, 'train')")))
        print(term.bold(term.magenta("2 . ") + term.green_on_black(
            "flow.comparing_statistics(dataframe_dict: dict)")))
        print(term.bold(term.magenta("For example: ") + term.green_on_black(
            "flow.comparing_statistics(dataframe_dict)")))
        print("\n")

    @staticmethod
    def target_based_categorical_feature_encoding():
        print(term.bold(term.magenta("Apply target-based encoding to the categorical features by applying the"
                                     " following function:\n")))
        print(term.bold(term.green_on_black(
            'dataframe_dict, columns_set = flow.features_encoding("target", dataframe_dict: dict, reference: str,'
            ' ignore_columns: list, target: str)')))
        print(term.magenta("An example of the ignore_columns list: \n") +
              term.green_on_black(" ignore_columns = ['id', 'target']\n"))
        print(term.magenta("An example of the reference: \n") +
              term.green_on_black(" reference = 'train'\n"))

    @staticmethod
    def frequency_based_categorical_feature_encoding():
        print(term.bold(term.magenta("Apply target-based encoding to the categorical features by applying the"
                                     " following function:\n")))
        print(term.bold(term.green_on_black(
            'dataframe_dict, columns_set = flow.features_encoding("frequency",'
            ' dataframe_dict: dict, reference: str, drop_encoded_features = False)')))
        print(term.magenta("An example of the ignore_columns list: \n") +
              term.green_on_black(" ignore_columns = ['id', 'target']\n"))
        print(term.magenta("An example of the reference: \n") +
              term.green_on_black(" reference = 'train'\n"))



import logging
import os
import pickle
import shutil
from typing import Union

import numpy as np
from sklearn.model_selection import train_test_split

logger = logging.getLogger(__name__)
formatting = (
    "%(asctime)s: %(levelname)s: File:%(filename)s Function:%(funcName)s Line:%(lineno)d "
    "message:%(message)s"
)
logging.basicConfig(
    filename=os.path.join(os.path.dirname(os.path.dirname(__file__)), "logs/logs.log"),
    level=logging.INFO,
    format=formatting,
)


def read_kfold_config(split: dict):
    """ KFold values reader

    This function ensures that the parameters of the KFold splitting method are defined.

    :param dict split: A dictionary that contains the parameters about the KFold splitting method.

    :return:
            - n_fold - An integer that refers to the number of folds which will be used for cross-validation.
            -  shuffle - A boolean. If true, data will be shuffled before splitting it to multiple folds.
            - random_state - An integer which helps to reproduce the results.

    """

    try:
        n_fold = split["fold_nr"]  # number of folds
    except Exception as e:
        print(f"fold_nr is not provided: {e}")
        n_fold = 5

    try:
        shuffle = split["shuffle"]
    except Exception as e:
        print(f"shuffle is not provided: {e}")
        shuffle = True

    try:
        random_state = split["random_state"]
    except Exception as e:
        print(f"random_state is not provided: {e}")
        random_state = 1

    return n_fold, shuffle, random_state


def create_model_directory(path: str):
    """ Model directory creator

    This function create a directory where the model during and after training will be saved.

    :param str path: It refers to the location where the models should be saved.

    """

    try:
        if not os.path.exists(path):
            os.makedirs(path)
            logger.info(f"{path} did not exist. New directory was created")
        else:
            shutil.rmtree(path)  # removes all the subdirectories!
            os.makedirs(path)
            logger.info(f"{path} existed. It was removed and new directory was created")
    except Exception as e:
        print(f"Creating models directory failed: {e}")
        logger.error(f"Creating models directory failed. Error is: {e}")


def save_model_locally(path: str, model: object):
    """ Model saver

    This function saves the model locally in pickle format.

    :param str path: It refers to the location where the models should be saved.
    :param object model: An object created by the training package e.g. Scikit Learn.

    """

    try:
        with open(path, "wb") as save_model:
            pickle.dump(model, save_model)
            logger.info(f"model was saved to {path}")
    except Exception as e:
        print(f"Saving model failed {e}")
        logger.error(f"Saving model failed. Error is: {e}")


def input_parameters_extraction(parameters: dict):
    """ Input data parsing

    :param parameters: dict parameters: A dictionary that contains information about the datasets, model type,
                model configurations and training configurations. Check the example below.
    :return:
            - data - A dictionary that contains pandas dataframes as datasets.
            - split - A dictionary that contains information about the cross-validation method.
            - train_array - A numpy array that is used to train the model and predict the target.
            - target - A numpy array that is used to train the model.
            - predict - If provided, a pandas dataframe that contains the features without the labels (target). Otherwise bool: False
    """

    try:
        data: dict = parameters["data"]
        split: dict = parameters["split"]

        train_array = data["train"]["features"].values
        target = np.array(data["train"]["target"])
        model_type = parameters["model"]["type"]
        hyperparameters = parameters["model"]["hyperparameters"]

        if "predict" in parameters.keys():
            predict: dict = parameters["predict"]
            logger.info("predict key is given")
        else:
            predict = False
            logger.info("predict key was not found")

        logger.info("Extracting the variables values is finished")
        return data, split, train_array, target, model_type, hyperparameters, predict

    except Exception as e:
        print(f"Parameters extraction failed: {e}")
        logger.error(f"Parameters extraction failed: {e}")


def split_dataset(features: np.array, target: np.array, tests_split_ratios: Union[list, set],
                  stratify: bool = False) -> dict:
    """ Dataset splitter

    This function split a dataset to multiple datasets such train, valid and test.

    :param np.array features: The original dataset that should be split to subsets
    :param np.array target: The original target/labels dataset that should be predicted
    :param Union[list, set] tests_split_ratios: A list or set of floats that represent the ratio of the size of the
            test dataset ot the train dataset. The values should be in the range ]0, 1[ e.g.
            tests_split_ratios = [0.2, 0.2]
    :param bool stratify: If set to True the ratios of the labels is kept the same in the splitted data sets.
    :return:
            sub_datasets: A dictionary that contains the test and train dataset

    :rtype: dict
    """

    # A counter for storing sub-datasets inside the dictionary
    split_count = 0
    sub_datasets = {}

    for split_ratio in tests_split_ratios:
        stratify_ = target if stratify else None
        features, x_test, target, y_test = train_test_split(features,
                                                            target,
                                                            test_size=split_ratio,
                                                            stratify=stratify_,
                                                            random_state=42)

        sub_datasets[f"features_{split_count}"] = x_test
        sub_datasets[f"target_{split_count}"] = y_test
        split_count += 1

    # The remaining dataset after having n split. This dataset will be used to train the model
    sub_datasets["train"] = features
    sub_datasets["target"] = target

    number_of_datasets = len(tests_split_ratios) + 1
    logger.info(f"Splitting the dataset to {number_of_datasets} sub-datasets is finished")
    return sub_datasets


def xgboost_problem_type(hyperparameters: dict) -> str:
    """ XGboost problem type finder

    The function finds the type of the problem that should be solved when using the xgboost method. The function uses
     the objective variable to find out what type of problem that should be solved

    :param dict hyperparameters: A dictionary that contains the hyperparameters which the selected training method
            needs to train the model.
    :return:
            - problem_to_solve: A string that defines the problem to solve: regression or classification.
    """

    try:
        # the default value of the objective is regression reg:squarederror.
        # Here the exception is caught to avoid that the objective is not given by the user
        objective = hyperparameters["objective"]
        if objective.startswith("reg"):
            problem_to_solve = "regression"
        else:
            problem_to_solve = "classification"
    except Exception as e:
        print(f"The objective is not defined. The default value is reg:squarederror. Error: {e}")
        problem_to_solve = "regression"
    return problem_to_solve


def define_model_directory_name(model_type: str, hyperparameters: dict, split: str, problem_to_solve: str):
    """ Assign directory name

    The function defines the name of the directory where the models will be saved locally after training.
    The name contains the information about the model type, split method and some information from the hyperparameters

    :param str model_type: The type of model that will be used to fit the data.
    :param dict hyperparameters: A dictionary that contains the hyperparameters which the selected training method
            needs to train the model.
    :param str split: A string that define the data splitting method: split or kfold
    :param str problem_to_solve: A string that defines the problem to solve: regression or classification.
    :return:
            - save_models_dir: The name of the directory where the trained models are saved locally
    """

    # define the name of the directory where the models will be saved
    if model_type == "Ridge linear regression":
        alpha = hyperparameters["alpha"]
        save_models_dir = os.path.join(".", "models", f'linear_ridge_{split}_{alpha}')
    elif model_type == "lightgbm":
        save_models_dir = os.path.join(
            ".", "models", f'lightgbm_{split}_{hyperparameters["num_leaves"]}' +
                           f'_{hyperparameters["boosting"]}')
    elif model_type == "Logistic regression":
        save_models_dir = os.path.join(".", "models", f'logistic_{split}')
    elif model_type == "xgboost":
        save_models_dir = os.path.join(".", "models", f'xgboost_{split}')
    elif model_type.split(".")[0] == "sklearn":
        save_models_dir = os.path.join(".", "models", model_type)
    else:
        raise ValueError("Model type is not recognized")

    save_models_dir += "_" + problem_to_solve

    return save_models_dir



import json
import logging
import os
from typing import Union

import pandas as pd
from pandas.io.json import json_normalize

logger = logging.getLogger(__name__)
formatting = (
    "%(asctime)s: %(levelname)s: File:%(filename)s Function:%(funcName)s Line:%(lineno)d "
    "message:%(message)s"
)
logging.basicConfig(
    filename=os.path.join(os.path.dirname(os.path.dirname(__file__)), "logs/logs.log"),
    level=logging.DEBUG,
    format=formatting,
)


def extract_json_from_list_dict(row_i: Union[list, dict], object_nr: int) -> dict:
    """ Valid JSON object extractor from list or dict

    The function extracts the valid JSON data directly from the list or the dict.

    :param Union[list, dict] row_i: The content of the row which has the index i in a Pandas series
    :param int object_nr: If there are multiple valid JSON objects detected, the object which has object_nr will be
            returned

    :return: valid_json_object that has the index object_nr
    :rtype: dict
    """

    if isinstance(row_i, list) and len(row_i) > object_nr:
        try:
            return json.loads(json.dumps(row_i[object_nr]))
        except Exception as e:
            logger.debug(f"Not valid JSON: {e}")
    elif isinstance(row_i, dict) and object_nr == 0:
        try:
            return json.loads(json.dumps(row_i))
        except Exception as e:
            logger.debug(f"Not valid JSON: {e}")
    return {}


def extract_json_objects(raw_string_data: str, start_json_object: list,
                         end_json_object: list, object_nr: int):
    """Valid JSON object extractor

    The function extracts valid JSON objects from texts. Those objects could not be extracted using the json.load

    :param str raw_string_data: The string object that could contain valid JSON objects
    :param list start_json_object: List of integers that point to the char "{" which can be the start of the JSON
            objects.
    :param list end_json_object: List of integers that point to the char "}" which can be the end of the JSON objects.
    :param int object_nr: points to the index of the object that should be extracted if there are multiple valid JSON
            objects.

    :return: valid_json_object that has the index object_nr
    :rtype: dict
    """

    # initiate variables
    counter = 0
    valid_json_objects = []
    current_position = 0

    for start_i in start_json_object:
        if start_i > current_position:
            for end_i in end_json_object:
                try:
                    valid_json_object = json.loads(raw_string_data[start_i:end_i + 1])
                    valid_json_objects.append(valid_json_object)
                    current_position = end_i
                    counter += 1
                    break
                except Exception as e:
                    logger.debug(f"Not valid JSON: {e}")
                    # not a valid JSON object
                    continue
        if counter == object_nr + 1:
            return [valid_json_objects[object_nr]]
    return {}


def normalize_feature(string_data: str, object_nr: int):
    """JSON data searcher

    This function searches for possible valid JSON data inside the given text. It identifies the possible JSON objects
    by defining their edges using "{" and "}". It passes each defined object to the "extract_json_objects" function
    to extract the valid JSON objects. The returned valid objects will be normalized and returned as a Pandas dataframe.

    :param str string_data: The string that could contain valid JSON objects.
    :param int object_nr: If there are multiple valid JSON objects detected, the object which has object_nr will be
            returned

    :return: Pandas dataframe which has "n" number of columns and one row.
    :rtype: pandas.DataFrame
    """

    json_valid_object = {}
    try:
        # This will raise an error if the string_data is nan
        if len(string_data) == 0:
            string_data = {}

        # Trying to normalize it directly if the input is not a string type
        json_valid_object = extract_json_from_list_dict(string_data, object_nr)

        # if the regular method fails to extract JSON, try the second
        if json_valid_object == {}:
            # Search for the start positions of the possible valid JSON objects
            start_json = [i for i, x in enumerate(string_data) if str(x) == "{"]

            # Search for the end positions of the possible valid JSON objects
            end_json = [i for i, x in enumerate(string_data) if str(x) == "}"]

            # Pass the string together with the start and end positions to the JSON data extractor
            json_valid_object = extract_json_objects(string_data, start_json, end_json, object_nr)

    except Exception as e:
        logger.debug(string_data)
        logger.debug(e)

    try:
        return json_normalize(json_valid_object)
    except Exception as e:
        logger.debug(f"not valid object to normalize: {json_valid_object}")
        logger.debug(f"the Error is {e}")
        return pd.DataFrame()


def apply_normalize_feature(dataseries: pd.Series, keys_amount: int):
    """JSON-dataframe converter

    :param pd.Series dataseries: the feature that contains possible JSON objects
    :param int keys_amount: The possible numbers of the keys or parent JSON object that a row may contain.

    :return: a list of dataframes. Each element of the this list represents the normalized JSON object in each row of
            the dataseries
    :rtype: list
    """

    # Bring together the JSON objects from each row.
    dataframe_list = []

    for key_i in range(keys_amount):
        # Bring together the JSON objects from one row
        df_list = []

        for row_i in range(dataseries.size):
            string_data = str(dataseries[row_i]). \
                replace("'", '"'). \
                replace("True", '"True"'). \
                replace("False", '"False"'). \
                replace("None", '"None"')
            normalized_object = normalize_feature(string_data, key_i)

            if normalized_object.shape[0] > 0:
                df_list.append(normalized_object)
            else:
                logger.debug("no object was found")

        df = pd.concat(df_list, axis=0, sort=False)
        logger.debug(f"the size of the dataframe from the object {key_i} is {df.shape}")

        # Change the name of the column by adding the key order
        columns = [x + f"_{key_i}" for x in df.columns]
        df.columns = columns
        logger.debug(columns)

        dataframe_list.append(df)
    return dataframe_list


def column_validation(dataframe: pd.DataFrame, parent_columns: list, feature: str):
    """Column name validator

    The function ensures that the dataframe doesn't have two features that have the same name. It changes the name of
    the column after normalizing the JSON object based on the name of the parent feature

    :param pd.DataFrame dataframe: the normalized JSON objects that were found in the given feature.
    :param list parent_columns: A list of the name of the features or columns of main dataset
    :param str feature: The name of the feature that contains the JSON objects

    :return: Pandas dataframe with valid names for the columns
    :rtype: pd.DataFrame
    """

    new_columns = []
    for column_i in dataframe.columns:

        # Give the name of the parent feature to the new extracted column
        column_i = feature + "_" + column_i

        # If the column already exists, add a suffix to the name of the column
        while column_i in parent_columns:
            column_i = column_i + "_1"

        new_columns.append(column_i)

    # Give the dataframe the valid names of the columns
    dataframe.columns = new_columns
    return dataframe


def combine_new_data_to_original(dataframe: pd.DataFrame, dataframe_list: list, feature: str):
    """Dataframes binder

    The function concatenates the original dataframe and the new created dataframe together.

    :param pd.DataFrame dataframe: the original dataframe / dataset
    :param list dataframe_list: list of the dataframes that are created from normalizing the JSON objects in each row
            of the given feature
    :param str feature: The name of the feature that contains JSON objects

    :return: Pandas dataframe that contains both the original and the new created datasets. The original feature will
            be deleted
    :rtype: pd.DataFrame
    """

    # First bring together the dataframes that contain the normalized objects from each row
    char_dataframe = pd.concat(dataframe_list, axis=1, sort=False)

    # Be sure that the index of the new created dataframe is correct and unique
    char_dataframe.index = range(char_dataframe.shape[0])

    # Be sure that the new dataframe has valid columns name (avoid two features have the same name)
    char_dataframe = column_validation(char_dataframe, dataframe.columns, feature)

    logger.debug(f"The shape of dataframe after normalizing the JSON data is: {char_dataframe.shape}")

    # Bring the old and the new datasets together
    joined_dataframe = pd.concat([dataframe, char_dataframe], axis=1, sort=False)

    # Dropping the original feature column after normalizing it.
    joined_dataframe = joined_dataframe.drop(feature, axis=1)

    return joined_dataframe


def feature_with_json_detector(dataseries: pd.Series):
    """ JSON detector

    This function detect if there is features in the dataset that could have valid JSON objects.

    :param pd.Series dataseries: The feature's values that should be tested for possible valid JSON objects

    :return: True if there is JSON objects candidates and False if not
    :rtype: bool
    """

    # Check which row contains start and end JSON syntax in case of string values inside the rows
    start_json = dataseries.fillna('').str.contains('{')
    end_json = dataseries.fillna('').str.contains('}')
    if start_json[start_json].size and end_json[end_json].size:
        for row_i in range(start_json.size):
            if start_json[row_i] and end_json[row_i]:
                raw_string_data = dataseries[row_i]

                # if the JSON is converted to a dataframe successfully, it is valid JSON
                if normalize_feature(raw_string_data, 0).shape[0]:
                    return True

    # Some of the rows can be valid lists that contain valid JSON format
    if start_json[start_json.isnull()].size and end_json[end_json.isnull()].size:
        for row_i in dataseries[start_json.isnull()]:
            for item_i in range(len(row_i)):
                if extract_json_from_list_dict(row_i, item_i) != {}:
                    return True

    return False


def combine_columns(dataframes_dict, feature):
    """
    For avoiding different numbers of generated columns for different datasets, I combine them in one large dataframe
    :param dataframes_dict:
    :param feature:
    :return:
    """
    dataframes_list = []
    combined_dataframe = pd.DataFrame()
    # I gave a key to easy recognize each dataset later.
    key_name = "key_dataframe"

    for key_i, dataframe in dataframes_dict.items():
        dataframe[key_name] = key_i
        dataframes_list.append(dataframe[[feature, key_name]])
        combined_dataframe = pd.concat(dataframes_list, axis=0, sort=False)
        combined_dataframe.index = range(combined_dataframe.shape[0])

    return combined_dataframe


def flat_json(dataframes_dict, json_columns, keys_amount=10):
    """
    :param dataframes_dict
    :param json_columns:
    :param keys_amount
    :return:
    """

    # Flatten JSON data in a certain feature
    for column_i in json_columns:

        dataframe_combined = combine_columns(dataframes_dict, column_i)

        dataseries = dataframe_combined[column_i]
        print(f"Flatting column: {column_i}")

        dataframe_list = apply_normalize_feature(dataseries, keys_amount)
        dataframe_combined = combine_new_data_to_original(dataframe_combined, dataframe_list, column_i)

        # Extract each data set and delete the added key and the original feature after it was flatten.
        for key_i in dataframes_dict.keys():
            extracted_dataframe = dataframe_combined[dataframe_combined["key_dataframe"] == key_i]

            # index should be reset otherwise there is a problem with the pd.concat
            extracted_dataframe.index = range(extracted_dataframe.shape[0])

            dataframes_dict[key_i] = pd.concat([dataframes_dict[key_i], extracted_dataframe], axis=1, sort=False)

            dataframes_dict[key_i] = dataframes_dict[key_i].drop([column_i, "key_dataframe"], axis=1)

    return dataframes_dict
